//
// Created by zygzagz on 05.01.2021.
//

#include <cassert>
#include <iostream>
#include <filesystem>
#include "Quadruple.h"
#include "../Compiler.h"
#include "QuadrupleGenerator.h"

vector<BasicBlockPtr> QuadrupleGenerator::compileFunction(FunctionInfoPtr f) {
	if (!f->block) return {};

	scope->currentFunction = f;
	block = make_shared<BasicBlock>();
	f->block->accept(this);

	int id = 1;
	for (auto v : vars) {
		if (v->constExpr && !v->info) {
			v->name = to_string(v->val);
			continue;
		}
		v->name = "@" + to_string(id);
		if (v->info) {
			v->name += "." + v->info->name + ":" + to_string(v->info->lineLocation);
		}
		id++;
	}

	flushBasicBlock();
	block = nullptr;

	for (auto b : blocks) {
		b->finishQuads();
	}

	int label = 1;
	for (auto b : blocks) {
		for (auto q : b->quads) {
			if (auto l = dynamic_pointer_cast<QLabel>(q)) {
				if (!l->label.empty())
					l->label = to_string(label++) + "_" + l->label;
				else
					l->label = to_string(label++);
			}
		}
	}

	for (auto v : vars) {
		if (v->info)
			v->info->loc = nullptr;
	}

	vars.clear();
	vector<BasicBlockPtr> ret;
	blocks.swap(ret);

	scope->currentFunction = nullptr;
	return ret;
}

/*
- kiedy przechodzę między blokami:
-	pierwsze wejście w while/fora: ustawienie używanych wartości w pętli w zależności od ostatnich wartości przed pętlą
-		dump przed wejściem w stmt fora, dump po sparsowaniu fora
-		różnice ustawić w bloku przed i tak samo za nim
-	if: przed ifem dump, za ifem dorobić else i ustawić diff
-	ifelse: przed ifem dump, w środku dump, po elsie dump
-		wszystkie poza zmienionymi w bloku ustawiamy na koniec bloku


 wskazówki do implementacji:
 - w momencie przypisywania do zmiennej lokalnej chcemy powiadomić najświeższy detektor zmian lokalnych
 - po zakończeniu bloku detektor zmian odpowiednio poprzypisuje w pożądanych miejcach odpowiednie definicje
 - po zakończeniu bloku detektor powiadomi kolejny z rzędu detektor o najnowszych instancjach zmienionych zmiennych
 - detektor powinien działać ponad blokami prostymi

 funkcje detektora:
 - dodanie zmiennej VarInfoPtr do zbioru, przyporządkowanie jej najnowszej definicji (i zapamiętanie oryginalnej)
 - umiejętność porównania dwóch zbiorów poprzez iterację
		- każdy element który występuje tylko w jednym
		- para jeśli el. występuje w obu zbiorach

implementacja detektora:
 - map<VarInfoPtr, pair<orig, new>>


 dalej: << i >> dla BB
 porobić graf BB
 złapać blocki i podumpować envy dookoła flow
 dopisać sumę dumpów
 */

BasicBlockPtr QuadrupleGenerator::flushBasicBlock() {
	auto ret = block;
	if (block) {
		blocks.emplace_back(block);
		flushLastQuad();
	}

	block = make_shared<BasicBlock>();

	return ret;
}

VariableLayout QuadrupleGenerator::captureEnv() {
	VariableLayout ret;
	BindingPtr b = scope->currentBinding;
	while (b) {
		for (const auto& info : b->variables) {
			if (info->isInstanceVariable()) continue;
			if (info->loc) {
				ret.add(info);
			}
		}
		b = b->getParent();
	}
	return ret;
}

/// expressions

void QuadrupleGenerator::visitEVar(EVar *p) {
	lastVar = alloc(p->pident_->var.lock());
}

void QuadrupleGenerator::visitELitInt(ELitInt *p) {
	lastVar = alloc(p->integer_);
}

void QuadrupleGenerator::visitELitTrue(ELitTrue *p) {
	lastVar = alloc(1);
}

void QuadrupleGenerator::visitELitFalse(ELitFalse *p) {
	lastVar = alloc(0);
}

void QuadrupleGenerator::visitEString(EString *p) {
	// TODO: zrobic liste stringow per app
	lastVar = alloc();
}

void QuadrupleGenerator::visitNullCast(NullCast *p) {
	lastVar = alloc(0);
}

void QuadrupleGenerator::visitNeg(Neg *p) {
	auto var = evalExpr(p->expr_);
	lastVar = alloc();
	addQuad<QAssign>(lastVar, Op::Neg, var);
}

void QuadrupleGenerator::visitNot(Not *p) {
	auto var = evalExpr(p->expr_);
	lastVar = alloc();
	addQuad<QAssign>(lastVar, Op::Not, var);
}

void QuadrupleGenerator::visitEMul(EMul *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();
	p->mulop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);
}

void QuadrupleGenerator::visitEAdd(EAdd *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();
	p->addop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);

}

void QuadrupleGenerator::visitERel(ERel *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();
	p->relop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);
}

void QuadrupleGenerator::visitEAnd(EAnd *p) {
	auto l = evalExpr(p->expr_1);
	auto labelLeft = make_shared<QLabel>("use_left");
	auto labelAfter = make_shared<QLabel>("end_and");

	addQuad<QJumpCond>(labelLeft, Op::Not, l);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc();

	addQuad<QAssign>(lastVar, Op::Copy, r);
	addQuad<QJump>(labelAfter);
	addQuad(labelLeft);
	addQuad<QAssign>(lastVar, Op::Copy, alloc(0));
	addQuad(labelAfter);
}

void QuadrupleGenerator::visitEOr(EOr *p) {
	auto l = evalExpr(p->expr_1);
	auto labelLeft = make_shared<QLabel>("use_left");
	auto labelAfter = make_shared<QLabel>("end_or");

	addQuad<QJumpCond>(labelLeft, Op::Copy, l);

	auto r = evalExpr(p->expr_2);
	lastVar = alloc();

	addQuad<QAssign>(lastVar, Op::Copy, r);
	addQuad<QJump>(labelAfter);
	addQuad(labelLeft);
	addQuad<QAssign>(lastVar, Op::Copy, alloc(1));
	addQuad(labelAfter);
}

/// complex extensions

void QuadrupleGenerator::visitEIndexAcc(EIndexAcc *p) {
	auto lhs = evalExpr(p->expr_1);
	auto index = evalExpr(p->expr_2);
	lastVar = alloc();
	auto type = dynamic_pointer_cast<Array>(lhs->type);
	if (!type) {
		throw ParseError("compiler evalExpr Variable has no Array type", p->expr_1);
	}
	if (!type->type_) {
		throw ParseError("compiler evalExpr Variable is Array of no type", p->expr_1);
	}
	if (type->type_->size() > 4) {
		// calculate quantifier, we multiply index because size is larger than 4
		auto q = alloc(type->type_->size() / 4);
		// calculate var = index * quantifier
		addQuad<QAssign>(lastVar, index, Op::Mul, q);

		// allocate new lastVar, index is now calculated multiplied index
		index = alloc();
		swap(index, lastVar);
	}
	auto quad = make_shared<QAccess>(lastVar, lhs, index, 4, 0);
	addQuad(quad);
}

void QuadrupleGenerator::visitEClsMmbr(EClsMmbr *p) {
	auto l = evalExpr(p->expr_);
	auto var = p->pident_->var.lock();
	assert(var);
	// it cannot be function, as function might only be used in EApp and are handled directly there
	assert(!var->toFunction());
	// Array length calculation works because of arrayLengthVar that has offset 0
	size_t offset = var->offset;
	if (dynamic_cast<EIndexAcc*>(p->expr_)) {
		// opt if EIndexAcc is used inside visitEClsMmbr
		auto access = dynamic_pointer_cast<QAccess>(block->quads.back());
		access->access.offset += offset;
		l->info = var;
		lastVar = l;
	} else {
		lastVar = alloc(var);
		auto quad = make_shared<QAccess>(lastVar, l, alloc(0), 0, offset);
		addQuad(quad);
	}
}

void QuadrupleGenerator::visitEApp(EApp *p) {
	auto funType = static_pointer_cast<Fun>(p->expr_->type);
	FunctionInfoPtr info;
	VariablePtr self;
	if (auto mem = dynamic_cast<EClsMmbr*>(p->expr_)) {
		// call tab[i].fun()
		// call obj.fun()
		self = evalExpr(mem->expr_);
		info = static_pointer_cast<FunctionInfo>(mem->pident_->var.lock());
	} else if (auto ident = dynamic_cast<EVar*>(p->expr_)) {
		auto klass = scope->currentFunction->klass.lock();
		info = static_pointer_cast<FunctionInfo>(ident->pident_->var.lock());
		auto otherClass = info->klass.lock();
		if (otherClass) {
			// call fun() = self.fun()
			assert(otherClass == klass);
			auto binding = scope->currentFunction->binding.lock();
			assert(binding);
			auto selfInfo = binding->variables.local("self");
			assert(selfInfo);
			self = alloc(selfInfo);
		} else {
			// call main()
		}
	} else {
		throw ParseError("Unimplemented EApp instantiation (neither EClsMmbr nor EVar", p);
	}

	int i = 1;
	for (auto param : *p->listexpr_) {
		auto var = evalExpr(param);
		addQuad<QParam>(var, i++);
	}

	if (self) {
		addQuad<QParam>(self, 0);
	}

	lastVar = alloc();
	addQuad<QCall>(lastVar, info, i, self);
}

void QuadrupleGenerator::visitENewArray(ENewArray *p) {
	// allocate enough memory with calloc, if it is class then init each entry...
	auto type = p->type_;
	size_t size = type->size();
	string virtSymbol;
	if (auto klass = dynamic_cast<ClassT*>(type)) {
		auto info = klass->pident_->var.lock()->toClass();
		size = info->size;
		virtSymbol = Compiler::getVirtName(info);
	}
	auto count = evalExpr(p->expr_);
	lastVar = alloc();
	addQuad<QAlloc>(lastVar, size, virtSymbol, count);
}

void QuadrupleGenerator::visitENewClass(ENewClass *p) {
	// allocate enough memory with calloc, setup virt table pointer.
	auto info = p->pident_->var.lock()->toClass();
	size_t size = info->size;
	string virtSymbol = Compiler::getVirtName(info);
	lastVar = alloc();
	addQuad<QAlloc>(lastVar, size, virtSymbol);
}

/// statements

void QuadrupleGenerator::visitInit(Init *p) {
	auto info = p->pident_->var.lock();
	assert(info);
	auto var = evalExpr(p->expr_);
	addQuad<QAssign>(alloc(info), Op::Copy, var);
}

void QuadrupleGenerator::assign(Expr* lval, VariablePtr val) {
	auto dest = evalLVal(lval);
	if (dest->info && dest->info->isInstanceVariable()) {
		// instance variable, need to write it to memory
		auto quad = dynamic_pointer_cast<QAccess>(lastQuad);
        assert(quad);
		auto loc = quad->access;
		lastQuad = nullptr;
		addQuad<QWrite>(loc, val);
	} else {
		// local variable - only assign
		addQuad<QAssign>(dest, Op::Copy, val);
	}
}

void QuadrupleGenerator::visitAss(Ass *p) {
	auto val = evalExpr(p->expr_2);
	assign(p->expr_1, val);
}

void QuadrupleGenerator::visitIncr(Incr *p) {
	auto lhs = evalExpr(p->expr_);
	auto tmp = alloc();
	addQuad<QAssign>(tmp, lhs, Op::Plus, alloc(1));
	assign(p->expr_, tmp);
}

void QuadrupleGenerator::visitDecr(Decr *p) {
	auto lhs = evalExpr(p->expr_);
	auto tmp = alloc();
	addQuad<QAssign>(tmp, lhs, Op::Minus, alloc(1));
	assign(p->expr_, tmp);
}

void QuadrupleGenerator::visitBlk(Blk *blk) {
	scope->currentBinding = blk->getBinding();

	Skeleton::visitBlk(blk);

	scope->currentBinding = scope->currentBinding->getParent();
}

/// flow control

void QuadrupleGenerator::visitRet(Ret *p) {
	auto var = evalExpr(p->expr_);
	addQuad<QReturn>(var);
	flushBasicBlock();
}

void QuadrupleGenerator::visitVRet(VRet *p) {
	addQuad<QReturn>(nullptr);
	flushBasicBlock();
}

void QuadrupleGenerator::visitCond(Cond *p) {
	compileCond(p->expr_, p->stmt_, nullptr);
}

void QuadrupleGenerator::visitCondElse(CondElse *p) {
	compileCond(p->expr_, p->stmt_1, p->stmt_2);
}

void QuadrupleGenerator::compileCond(Expr *expr_, Stmt *stmt_1, Stmt *stmt_2) {
	auto elseBranch = make_shared<QLabel>("else");
	auto after = make_shared<QLabel>("end_else");

	auto var = evalExpr(expr_);

	addQuad<QJumpCond>(elseBranch, Op::Not, var);
	auto beforeBlock = flushBasicBlock(); // possible jump -> else
	auto env1 = captureEnv(), env2 = env1;

	stmt_1->accept(this);
	addQuad<QJump>(after);
	auto stmt1Block = flushBasicBlock(); // jump -> after
	env1.capture();
	env1.revert();

	addQuad(elseBranch);
	if (stmt_2) {
		stmt_2->accept(this);
	}

	auto stmt2Block = flushBasicBlock(); // jump <- cond
	env2.capture();
	addQuad(after);

	beforeBlock->append(stmt1Block);
	beforeBlock->append(stmt2Block);
	stmt1Block->append(block);
	stmt2Block->append(block);

	merge2Envs(&env1, stmt1Block, &env2, stmt2Block);
}

void QuadrupleGenerator::merge2Envs(VariableLayout *env1, BasicBlockPtr b1, VariableLayout *env2, BasicBlockPtr b2) {
	for (int i = 0;  i < 2; i++) {
		for (const auto& p1 : env1->changes) {
			if (p1.second.first == p1.second.second) continue;
			auto info = p1.first;
			auto it = env2->changes.find(info);
			if (it != env2->changes.end()) {
				auto p2 = *it;
				// variables have been already merged
				if (p1.second.first == p2.second.second) continue;
				info->loc = nullptr;
				auto merged = alloc(info);
				// assign both versions to merged variable location
				b1->addJumpInitQuad(make_shared<QAssign>(merged, Op::Copy, p1.second.second));
				b2->addJumpInitQuad(make_shared<QAssign>(merged, Op::Copy, p2.second.second));
			} else {
				// copy changes to another block that did not touch var
				b2->addJumpInitQuad(make_shared<QAssign>(p1.second.second, Op::Copy, p1.second.first));
			}
		}

		swap(env1, env2);
		swap(b1, b2);
	}
}

void QuadrupleGenerator::visitWhile(While *expr) {
	auto cond = make_shared<QLabel>("cond");

	addQuad<QJump>(cond);
	auto beforeBlock = flushBasicBlock(); // jump <- loop -> cond
	auto loop = make_shared<QLabel>("loop");
	addQuad(loop); // jump <- loop

	// hook env
	auto env1 = captureEnv();

	// produce all variables hooks
	for (const auto& change : env1.changes) {
		auto info = change.first;
		info->loc = nullptr;
		alloc(info);
	}

	// save hooks for later
	env1.capture();
	// env2 starts with hooked variables
	auto env2 = captureEnv();

	expr->stmt_->accept(this);

	auto loopBlock = flushBasicBlock(); // jump <- cond
	addQuad(cond);

	// env2 contains changed hooks
	env2.capture();
	// restore env1 pre-hook variables
	env1.revert();

	for (const auto& p : env2.changes) {
		auto info = p.first;
		auto x1 = p.second.first;
		auto x2 = p.second.second;
		// save hooks if used
		info->loc = x1;
		// transition from pre-hook to hooked var [before -> cond]
		beforeBlock->addJumpInitQuad(make_shared<QAssign>(x1, Op::Copy, env1.changes[info].first));
		// transition from loop var to hooked var [loop -> cond]
		loopBlock->addJumpInitQuad(make_shared<QAssign>(x1, Op::Copy, x2));
	}

	// expr uses pre-hook variables iff unused in loop
	auto var = evalExpr(expr->expr_);

	addQuad<QJumpCond>(loop, Op::Copy, var);
	auto condBlock = flushBasicBlock(); // jump -> loop

	// next block is ready to use updated variables

	beforeBlock->append(condBlock);
	condBlock->append(loopBlock);
	condBlock->append(block);
}

void QuadrupleGenerator::visitSExp(SExp *p) {
	evalExpr(p->expr_);
}

void QuadrupleGenerator::visitForEach(ForEach *p) {
	auto tab = evalExpr(p->expr_);
	throw ParseError("ForEach is yet unimplemented!", p);
	// TODO: implement
}